from typing import Text, Dict, Any, List
import hashlib
import json
import ctypes
import multiprocessing
import asyncio
import uuid
import socket
from unittest import mock

from rasax.community.telemetry import (
    REPOSITORY_CREATED_EVENT,
    GIT_CHANGES_PUSHED_EVENT,
    FEATURE_FLAG_UPDATED_EVENT,
)
from sqlalchemy.orm import Session
import pytest
from _pytest.monkeypatch import MonkeyPatch
from aioresponses import aioresponses, CallbackResult
from rasa.nlu.training_data.formats import MarkdownReader
from rasax.community import config, telemetry
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.story_service import StoryService
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.model_service import ModelService
from rasax.community.services.event_service import EventService
from rasax.community.services.config_service import ConfigService, ConfigKey
from rasax.community.services.settings_service import SettingsService
from rasax.community.constants import RASA_PRODUCTION_ENVIRONMENT
from rasax.community.version import __version__
from tests.unit.conftest import insert_test_model, user_event, reset_config_service
from tests.unit.services.test_domain_service import TEST_DOMAIN
from tests.unit.services.test_data_service import EXAMPLE_DATA_MD
from tests.unit.services.test_story_service import TEST_STORY
from tests.unit.services.test_settings_service import default_environments_config

TELEMETRY_TEST_USER = uuid.uuid4().hex


def _mock_track_internal_exception(*args, **kwargs) -> None:
    raise Exception("Tracking failed")


@pytest.fixture()
def clean_config_service(config_service: ConfigService) -> None:
    # Clean before testing
    reset_config_service(config_service)
    yield
    # Clean up after testing (in case other tests use the service without resetting it)
    reset_config_service(config_service)


def test_get_telemetry_config(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(telemetry, "_telemetry_enabled", None)
    monkeypatch.setattr(telemetry, "_telemetry_id", None)

    # If config variables weren't initialized, they should be after we retrieve them
    telemetry_enabled, telemetry_id = telemetry.get_telemetry_config()

    assert telemetry_enabled is telemetry._telemetry_enabled
    assert telemetry_id is telemetry._telemetry_id

    assert not telemetry_enabled.value
    assert telemetry_id.value == b""

    # If we retrieve them again, we should get the same values (by identity)
    telemetry_enabled_bis, telemetry_id_bis = telemetry.get_telemetry_config()
    assert telemetry_enabled is telemetry_enabled_bis
    assert telemetry_id is telemetry_id_bis


def test_initialize_configuration_from_file_server_mode(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(config, "LOCAL_MODE", False)

    telemetry.initialize_configuration_from_file(None)
    assert telemetry._telemetry_enabled is None
    assert telemetry._telemetry_id is None


def test_initialize_configuration_from_database_local_mode(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(config, "LOCAL_MODE", True)

    telemetry.initialize_configuration_from_db(None)
    assert telemetry._telemetry_enabled is None
    assert telemetry._telemetry_id is None


@pytest.fixture()
def reset_telemetry_config(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(
        telemetry, "_telemetry_enabled", multiprocessing.Value(ctypes.c_bool, False)
    )

    monkeypatch.setattr(
        telemetry,
        "_telemetry_id",
        multiprocessing.Array(ctypes.c_char, telemetry.TELEMETRY_ID_LENGTH),
    )


def test_initialize_configuration_from_database_clean(
    clean_config_service: None,
    reset_telemetry_config: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Clean config: no environment variables and default config values in DB
    # Telemetry should be enabled

    monkeypatch.delenv(ConfigKey.TELEMETRY_ENABLED, raising=False)
    monkeypatch.delenv(ConfigKey.TELEMETRY_UUID, raising=False)

    telemetry.initialize_configuration_from_db(session)

    assert telemetry.is_telemetry_enabled()
    assert telemetry.telemetry_id() == config_service.get_value(
        ConfigKey.TELEMETRY_UUID
    )


def test_initialize_configuration_from_database_disabled(
    clean_config_service: None,
    reset_telemetry_config: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Telemetry has been disabled by setting the configuration option in the database
    # No env variables
    # Telemetry should be disabled

    config_service.set_value(ConfigKey.TELEMETRY_ENABLED, False)

    monkeypatch.delenv(ConfigKey.TELEMETRY_ENABLED, raising=False)

    telemetry.initialize_configuration_from_db(session)

    assert not telemetry.is_telemetry_enabled()


def test_initialize_configuration_from_database_env_enabled(
    clean_config_service: None,
    reset_telemetry_config: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Telemetry has been enabled via environment (in most cases: upgrading
    # from 0.23 and using same execution context to start the server)
    # New Rasa X 0.24 installs should not configure telemetry via environment!
    # Telemetry should be enabled

    telemetry_id = uuid.uuid4().hex

    monkeypatch.setenv(ConfigKey.TELEMETRY_ENABLED, "true")
    monkeypatch.setenv(ConfigKey.TELEMETRY_UUID, telemetry_id)

    telemetry.initialize_configuration_from_db(session)

    assert telemetry.is_telemetry_enabled()
    assert telemetry.telemetry_id() == telemetry_id


@pytest.mark.parametrize("telemetry_consent", ["", "false", "foobar"])
def test_initialize_configuration_from_database_env_disabled(
    telemetry_consent: Text,
    clean_config_service: None,
    reset_telemetry_config: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Same test as previous one, but telemetry has been explicitly disabled via
    # environment variables.
    # Telemetry should be disabled

    monkeypatch.setenv(ConfigKey.TELEMETRY_ENABLED, telemetry_consent)

    telemetry.initialize_configuration_from_db(session)

    assert not telemetry.is_telemetry_enabled()


async def test_track_async(enable_telemetry: None, monkeypatch: MonkeyPatch):
    calls = 0

    def _response(url, **kwargs):
        nonlocal calls
        calls += 1
        return CallbackResult(body=json.dumps({"success": True}))

    # If we're running on Travis CI, then no POST requests should be sent
    monkeypatch.setenv("TRAVIS", "true")
    monkeypatch.setenv("CI", "true")

    with aioresponses() as responses:
        responses.post(telemetry.SEGMENT_ENDPOINT, callback=_response)
        await telemetry._track_async(TELEMETRY_TEST_USER, "Test", {"foo": "bar"})

    assert calls == 0  # No actual POST requests done

    # If not on CI, then a POST request should have been made
    monkeypatch.delenv("TRAVIS")
    monkeypatch.delenv("CI")

    with aioresponses() as responses:
        responses.post(telemetry.SEGMENT_ENDPOINT, callback=_response)
        await telemetry._track_async(TELEMETRY_TEST_USER, "Test", {"foo": "bar"})

    assert calls == 1

    # If the request triggers an exception, the exception should be logged and
    # not propagated
    with aioresponses() as responses:
        responses.post(telemetry.SEGMENT_ENDPOINT, exception=socket.gaierror())
        await telemetry._track_async(TELEMETRY_TEST_USER, "Test", {"foo": "bar"})
        # If no exception was raised, the test has passed


def test_track_internal_loop_stopped(enable_telemetry: None, monkeypatch: MonkeyPatch):
    # Test that _track_internal works correctly when there's no current event
    # loop running.
    magic_mock = mock.MagicMock()

    @asyncio.coroutine
    def _mock_track_async(*args, **kwargs):
        magic_mock(*args, **kwargs)

    monkeypatch.setattr(telemetry, "_track_async", _mock_track_async)
    loop = asyncio.get_event_loop()
    assert not loop.is_running()

    telemetry.track("Test", {"foo": "bar"}, {"quux": "baz"})

    assert magic_mock.call_count == 1
    assert magic_mock.call_args[0] == (
        TELEMETRY_TEST_USER,
        "Test",
        {"foo": "bar", telemetry.TELEMETRY_ID: TELEMETRY_TEST_USER},
        {"quux": "baz"},
    )


async def test_track_internal_loop_running(
    enable_telemetry: None, monkeypatch: MonkeyPatch
):
    # Test that _track_internal works correctly when there's an event loop
    # already running.
    magic_mock = mock.MagicMock()

    @asyncio.coroutine
    def _mock_track_async(*args, **kwargs):
        magic_mock(*args, **kwargs)

    monkeypatch.setattr(telemetry, "_track_async", _mock_track_async)
    loop = asyncio.get_event_loop()
    assert loop.is_running()

    telemetry.track("Test", {"foo": "bar"}, {"quux": "baz"})

    # Since _track_internal() is fire-and-forget when a global event loop is
    # running, wait a second before running the assertions to make sure the
    # funcion has been executed (and therefore the mock used).
    await asyncio.sleep(1)

    assert magic_mock.call_count == 1
    assert magic_mock.call_args[0] == (
        TELEMETRY_TEST_USER,
        "Test",
        {"foo": "bar", telemetry.TELEMETRY_ID: TELEMETRY_TEST_USER},
        {"quux": "baz"},
    )


def test_track_ignore_exception(enable_telemetry: None, monkeypatch: MonkeyPatch):
    monkeypatch.setattr(telemetry, "_track_internal", _mock_track_internal_exception)

    # If the test finishes without raising any exceptions, then it's successful
    telemetry.track("Test")


def test_get_project_status_event_empty(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
    new_project: Text,
):
    actual = telemetry._get_project_status_event(session, new_project)

    assert actual.pop("uptime_seconds") > 0
    assert actual == {
        "project": hashlib.sha256(new_project.encode("utf-8")).hexdigest(),
        "local_mode": False,
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 0,
        "num_templates": 0,
        "num_slots": 0,
        "num_forms": 0,
        "num_intents": 0,
        "num_entities": 0,
        "num_stories": 0,
        "num_conversations": 0,
        "num_models": 0,
        "num_lookup_table_files": 0,
        "num_lookup_table_entries": 0,
        "num_synonyms": 0,
        "num_regexes": 0,
        "num_environments": 0,
        "environment_names": [],
        "num_live_environments": 0,
    }


async def test_get_project_status_event(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
    nlg_service: NlgService,
    event_service: EventService,
    new_project: Text,
    user: Dict[Text, Any],
):
    domain_service.store_domain(TEST_DOMAIN, new_project)
    domain_id = domain_service.get_domain_id(new_project)
    assert domain_id is not None

    nlg_service.replace_templates(
        [{"template": "utter_greet", "text": "Hi there!"}], user["username"], domain_id
    )

    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    model_service = ModelService("path", session)
    insert_test_model(new_project, "telemetry_test_model", model_service)

    await story_service.save_stories(TEST_STORY, user["team"])

    test_event = user_event("telemetry_test_user")
    event_service.save_event(json.dumps(test_event))

    settings_service = SettingsService(session)
    settings_service.save_environments_config(
        new_project, default_environments_config()["environments"]
    )

    session.commit()

    actual = telemetry._get_project_status_event(session, new_project)
    actual["environment_names"].sort()

    assert actual.pop("uptime_seconds") > 0
    assert actual == {
        "project": hashlib.sha256(new_project.encode("utf-8")).hexdigest(),
        "local_mode": False,
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 1,
        "num_templates": 1,
        "num_slots": 3,
        "num_forms": 1,
        "num_intents": 4,
        "num_entities": 2,
        "num_stories": 2,
        "num_conversations": 1,
        "num_models": 1,
        "num_lookup_table_files": 1,
        "num_lookup_table_entries": 7,
        "num_synonyms": 2,
        "num_regexes": 1,
        "num_environments": 3,
        "environment_names": ["development", RASA_PRODUCTION_ENVIRONMENT, "worker"],
        "num_live_environments": 0,
    }


@pytest.mark.parametrize(
    "referrer, origin",
    [
        ("http://localhost/interactive", "interactive"),
        ("http://localhost/stories", "stories"),
    ],
)
def test_track_story_created(telemetry_events: List, referrer: Text, origin: Text):
    telemetry.track_story_created(referrer)

    assert telemetry_events[0] == {
        "event": "Story Created",
        "properties": {"story_created_from": origin},
        "context": None,
    }


@pytest.mark.parametrize(
    "username, input_channel",
    [
        ("me", "rasa"),
        ("test-user", "rasa"),
        ("3d1693ea-9c51-4410-9e31-444e8a35727d", "Tester"),
        ("3d1693ea-9c51-4410-9e31-444e8a35727d", None),
    ],
)
def test_track_message_received(
    telemetry_events: List, username: Text, input_channel: Text
):
    telemetry.track_message_received(username, input_channel)

    assert telemetry_events[0] == {
        "event": "Message Received",
        "properties": {
            "username": hashlib.sha256(username.encode("utf-8")).hexdigest(),
            "channel": input_channel or "rasa",
        },
        "context": None,
    }


@pytest.mark.parametrize(
    "referrer, origin",
    [
        ("http://localhost/conversations", "conversations"),
        (
            "http://localhost/conversations/bcf7bb85-0506-42fb-89d4-5c1e310c4133/13.16",
            "conversations",
        ),
        ("http://localhost/data", "annotate_new_data"),
    ],
)
def test_track_message_annotated_from_referrer(
    telemetry_events: List, referrer: Text, origin: Text
):
    telemetry.track_message_annotated_from_referrer(referrer)

    assert telemetry_events[0] == {
        "event": "Message Annotated",
        "properties": {"message_annotated_from": origin},
        "context": None,
    }


def test_track_message_annotated(telemetry_events: List):
    telemetry.track_message_annotated(telemetry.MESSAGE_ANNOTATED_CONVERSATIONS)

    assert telemetry_events[0] == {
        "event": "Message Annotated",
        "properties": {
            "message_annotated_from": telemetry.MESSAGE_ANNOTATED_CONVERSATIONS
        },
        "context": None,
    }


def test_track_telemetry_disabled(telemetry_events: List, monkeypatch: MonkeyPatch):
    # Telemetry ID present but consent is False
    telemetry._telemetry_enabled.value = False
    telemetry._telemetry_id.value = uuid.uuid4().hex.encode("utf-8")

    telemetry.track("Foobar")
    assert len(telemetry_events) == 0


def test_track_telemetry_without_user(telemetry_events: List):
    # Telemetry consent is True but no ID present
    telemetry._telemetry_enabled.value = True
    telemetry._telemetry_id.value = bytes()

    telemetry.track("Foobar")

    assert len(telemetry_events) == 0


def test_track_repository_creation(telemetry_events: List):
    test_branch_name = "test_track_repository_created"
    telemetry.track_repository_creation(test_branch_name)

    assert telemetry_events[0] == {
        "event": REPOSITORY_CREATED_EVENT,
        "properties": {"branch": test_branch_name},
        "context": None,
    }


def test_git_changes_pushed(telemetry_events: List):
    test_branch_name = "test_git_changes_pushed"
    telemetry.track_git_changes_pushed(test_branch_name)

    assert telemetry_events[0] == {
        "event": GIT_CHANGES_PUSHED_EVENT,
        "properties": {"branch": test_branch_name},
        "context": None,
    }


@pytest.mark.parametrize("enabled", [True, False])
def test_track_feature_flag(telemetry_events: List, enabled: bool):
    feature_name = "test_track_feature_flag"

    telemetry.track_feature_flag(feature_name, enabled)

    assert telemetry_events[0] == {
        "event": FEATURE_FLAG_UPDATED_EVENT,
        "properties": {"name": feature_name, "enabled": enabled},
        "context": None,
    }
